#include "NoSelEOL.h"


/* >>>>>>>>>>>>>>>>>>>>>>>> selection state >>>>>>>>>>>>>>>>>>>>>>>> */
typedef struct tSelectionState {
    BOOL      bSelChange;
    BOOL      bSelJustChanged;
    BOOL      bMouseLButtonDown;
    BOOL      bShiftDown;
    BOOL      bColumnSelection;
    CHARRANGE SelKeyDownRange;
} SelectionState;

BOOL isSelectionChanging(const SelectionState* pSelState)
{
    if ( pSelState->bSelChange )
    {
        if ( pSelState->bMouseLButtonDown ||
             pSelState->bShiftDown )
        {
            return TRUE;
        }
    }
    return FALSE;
}

BOOL isSelectionChanged(const SelectionState* pSelState)
{
    if ( pSelState->bSelChange )
    {
        if ( (!pSelState->bMouseLButtonDown) &&
             (!pSelState->bShiftDown) )
        {
            return TRUE;
        }
    }
    return FALSE;
}

void clearSelectionState(SelectionState* pSelState)
{
    pSelState->bSelChange = FALSE;
    pSelState->bSelJustChanged = FALSE;
    pSelState->bMouseLButtonDown = FALSE;
    pSelState->bShiftDown = FALSE;
    pSelState->bColumnSelection = FALSE;
    pSelState->SelKeyDownRange.cpMin = 0;
    pSelState->SelKeyDownRange.cpMax = 0;
}

void initializeSelectionState(SelectionState* pSelState)
{
    clearSelectionState(pSelState);
}
/* <<<<<<<<<<<<<<<<<<<<<<<< selection state <<<<<<<<<<<<<<<<<<<<<<<< */


/* >>>>>>>>>>>>>>>>>>>>>>>> plugin state >>>>>>>>>>>>>>>>>>>>>>>> */
typedef struct tPluginState {
    DWORD        dwInitializedMode;
    HWND         hMainWnd;
    BOOL         bOldWindows;
    BOOL         bOldRichEdit; // TRUE means Rich Edit 2.0
    BOOL         bMDI;
    BOOL         bAkelEdit;
    WNDPROCDATA* pMainProcData;
    WNDPROCDATA* pEditProcData;
    WNDPROCDATA* pFrameProcData;
} PluginState;

void initializePluginState(PluginState* pPlugin)
{
    pPlugin->dwInitializedMode = 0;
    pPlugin->bAkelEdit = FALSE;
    pPlugin->hMainWnd = NULL;
    pPlugin->pMainProcData = NULL;
    pPlugin->pEditProcData = NULL;
    pPlugin->pFrameProcData = NULL;
}
/* <<<<<<<<<<<<<<<<<<<<<<<< plugin state <<<<<<<<<<<<<<<<<<<<<<<< */


/* >>>>>>>>>>>>>>>>>>>>>>>> data buffer >>>>>>>>>>>>>>>>>>>>>>>> */
typedef struct tDataBuffer {
    void* pData;      // pointer to the data
    int   nDataSize;  // data size in bytes
} DataBuffer;

void bufInitialize(DataBuffer* pBuf)
{
    pBuf->pData = NULL;
    pBuf->nDataSize = 0;
}

void bufFree(DataBuffer* pBuf)
{
    if ( pBuf->pData )
    {
        GlobalFree( (HGLOBAL) pBuf->pData );
        pBuf->pData = NULL;
    }
    pBuf->nDataSize = 0;
}

BOOL bufReserve(DataBuffer* pBuf, int nBytesToReserve)
{
    if ( pBuf->nDataSize < nBytesToReserve )
    {
        bufFree(pBuf);

        nBytesToReserve = (1 + nBytesToReserve/64)*64;

        if ( pBuf->pData = GlobalAlloc(GMEM_FIXED, nBytesToReserve) )
            pBuf->nDataSize = nBytesToReserve;
        else
            return FALSE;
    }
    return TRUE;
}
/* <<<<<<<<<<<<<<<<<<<<<<<< data buffer <<<<<<<<<<<<<<<<<<<<<<<< */


// consts
const char* cszPluginName = "NoSelEOL Plugin";
const char* cszUnloadPluginQuestion = "Plugin is active. Unload it?";


// vars
PluginState    g_Plugin;
SelectionState g_SelState;
DataBuffer     g_TextBuf;
BOOL           g_bSmartEndMode2 = FALSE;
/*BOOL           g_bProcessEndKey = TRUE;*/
BOOL           g_bDelLine = FALSE;


#define  MODE_NOSELEOL  0x01
#define  MODE_SMARTEND  0x02

// funcs
void Initialize(PLUGINDATA* pd, DWORD dwMode);
void Uninitialize(DWORD dwMode);
LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void OnEditSelKeyDown(HWND hEdit, unsigned int uKey);
void OnEditSelKeyUp(HWND hEdit, unsigned int uKey);
void OnEditSelChange(HWND hEdit);
void OnEditSelectionChanged(HWND hEdit);
BOOL OnEditEndKeyDown(HWND hEdit, LPARAM lParam);
void CheckEditNotification(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);


// Identification
/* extern "C" */
void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
    pv->dwAkelDllVersion  = AKELDLL;
    pv->dwExeMinVersion3x = MAKE_IDENTIFIER(3, 7, 4, 0);
    pv->dwExeMinVersion4x = MAKE_IDENTIFIER(4, 4, 0, 0);
    pv->pPluginName = "NoSelEOL";
}

// Plugin extern function
/* extern "C" */ 
void __declspec(dllexport) Main(PLUGINDATA *pd)
{
    if ( pd->lpbAutoLoad )
    {
        *(pd->lpbAutoLoad) = TRUE;
        return;
    }

    // Is plugin already loaded?
    if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
    {
        Uninitialize(MODE_NOSELEOL);
        pd->nUnload = UD_NONUNLOAD_NONACTIVE;
        return;
    }
    else
    {
        Initialize(pd, MODE_NOSELEOL);
    }
  
    // Stay in memory, and show as active
    pd->nUnload = UD_NONUNLOAD_ACTIVE;
}

// Plugin extern function
/* extern "C" */ 
void __declspec(dllexport) SmartEnd(PLUGINDATA *pd)
{
    if ( pd->lpbAutoLoad )
    {
        *(pd->lpbAutoLoad) = TRUE;
        return;
    }

    // Is plugin already loaded?
    if ( g_Plugin.dwInitializedMode & MODE_SMARTEND )
    {
        Uninitialize(MODE_SMARTEND);
        pd->nUnload = UD_NONUNLOAD_NONACTIVE;
        return;
    }
    else
    {
        Initialize(pd, MODE_SMARTEND);
    }
  
    // Stay in memory, and show as active
    pd->nUnload = UD_NONUNLOAD_ACTIVE;
}

// Plugin extern function
/* extern "C" */ 
void __declspec(dllexport) altSmartEnd(PLUGINDATA *pd)
{
    if ( pd->lpbAutoLoad )
    {
        *(pd->lpbAutoLoad) = TRUE;
        return;
    }

    g_bSmartEndMode2 = !g_bSmartEndMode2;
    
    pd->nUnload = g_bSmartEndMode2 ? UD_NONUNLOAD_ACTIVE : UD_NONUNLOAD_NONACTIVE;
}

// Entry point
/* extern "C" */ 
BOOL WINAPI DllMain(HINSTANCE hInstDLL, DWORD dwReason, LPVOID lpReserved)
{
    switch ( dwReason )
    {
        case DLL_PROCESS_ATTACH:
            initializePluginState(&g_Plugin);
            initializeSelectionState(&g_SelState);
            bufInitialize(&g_TextBuf);
            break;
        case DLL_PROCESS_DETACH:
            Uninitialize(MODE_NOSELEOL | MODE_SMARTEND);
            break;
        case DLL_THREAD_ATTACH:
            break;
        case DLL_THREAD_DETACH:
            break;
        default:
            break;
    }
    return TRUE;
}

void Initialize(PLUGINDATA* pd, DWORD dwMode)
{
    g_Plugin.hMainWnd     = pd->hMainWnd;
    g_Plugin.bOldWindows  = pd->bOldWindows;
    g_Plugin.bOldRichEdit = pd->bOldRichEdit;
    g_Plugin.bMDI         = pd->bMDI;
    g_Plugin.bAkelEdit    = pd->bAkelEdit;

    g_bDelLine = FALSE;

    // dwMode mask (depends on possible mode flags)
    dwMode &= (MODE_NOSELEOL | MODE_SMARTEND);
    
    if ( dwMode & MODE_NOSELEOL )
    {
        // this SubClassing is required 
        // only for MODE_NOSELEOL 
        if ( (g_Plugin.dwInitializedMode & MODE_NOSELEOL) == 0 )
        {
            clearSelectionState(&g_SelState);

            // SubClass main window
            g_Plugin.pMainProcData = NULL;
            SendMessage( g_Plugin.hMainWnd, AKD_SETMAINPROC,
                (WPARAM) NewMainProc, (LPARAM) &g_Plugin.pMainProcData );

            if ( g_Plugin.bMDI )
            {
                // SubClass frame window
                g_Plugin.pFrameProcData = NULL;
                SendMessage( g_Plugin.hMainWnd, AKD_SETFRAMEPROC,
                    (WPARAM) NewFrameProc, (LPARAM) &g_Plugin.pFrameProcData );
            }
        }
    }
  
    if ( dwMode )
    {
        // this SubClassing is required 
        // for both MODE_NOSELEOL and MODE_SMARTEND
        if ( !g_Plugin.dwInitializedMode )
        {
            // SubClass edit window
            g_Plugin.pEditProcData = NULL;
            SendMessage( g_Plugin.hMainWnd, AKD_SETEDITPROC,
                (WPARAM) NewEditProc, (LPARAM) &g_Plugin.pEditProcData );
        }
    }

    g_Plugin.dwInitializedMode |= dwMode;
}

void Uninitialize(DWORD dwMode)
{
    if ( !g_Plugin.dwInitializedMode )
        return;

    // dwMode mask (depends on current value of dwInitializedMode)
    dwMode &= g_Plugin.dwInitializedMode;

    if ( dwMode & MODE_NOSELEOL )
    {
        // these functions were subclassed 
        // only for MODE_NOSELEOL
        if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
        {
            if ( g_Plugin.pMainProcData )
            {
                // Remove subclass (main window)
                SendMessage( g_Plugin.hMainWnd, AKD_SETMAINPROC, 
                    (WPARAM) NULL, (LPARAM) &g_Plugin.pMainProcData );
                g_Plugin.pMainProcData = NULL;
            }
            if ( g_Plugin.bMDI && g_Plugin.pFrameProcData )
            {
                // Remove subclass (frame window)
                SendMessage( g_Plugin.hMainWnd, AKD_SETFRAMEPROC,
                    (WPARAM) NULL, (LPARAM) &g_Plugin.pFrameProcData );
                g_Plugin.pFrameProcData = NULL;
            }
        }
    }

    if ( dwMode )
    {
        if ( g_Plugin.dwInitializedMode & MODE_SMARTEND )
        {
            if ( dwMode & MODE_SMARTEND )
            {
                bufFree(&g_TextBuf);
            }
        }
        // this function was subclassed 
        // for both MODE_NOSELEOL and MODE_SMARTEND
        if ( g_Plugin.dwInitializedMode == dwMode ) // dwMode is already masked
        {
            if ( g_Plugin.pEditProcData )
            {
                // Remove subclass (edit window)
                SendMessage( g_Plugin.hMainWnd, AKD_SETEDITPROC,
                    (WPARAM) NULL, (LPARAM) &g_Plugin.pEditProcData );
                g_Plugin.pEditProcData = NULL;
            }
        }
    }

    g_Plugin.dwInitializedMode -= dwMode;
}

LRESULT CALLBACK NewEditProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch ( uMsg )
    {
        case WM_KEYDOWN:
        {
            LRESULT lResult = 0;

            if ( g_Plugin.dwInitializedMode & MODE_SMARTEND )
            {
                if ( (wParam == VK_END) /*&& g_bProcessEndKey*/ )
                {
                    if ( ((GetKeyState(VK_CONTROL) & 0x80) == 0) &&
                         ((GetKeyState(VK_MENU) & 0x80) == 0) )
                    {
                        if ( OnEditEndKeyDown(hWnd, lParam) )
                            return TRUE;
                    }
                }
            }
            else if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                if ( (wParam == VK_END) /*&& g_bProcessEndKey*/ )
                {
                    if ( GetKeyState(VK_SHIFT) & 0x80 )
                    {
                        if ( OnEditEndKeyDown(hWnd, lParam) )
                            return TRUE;
                    }
                }
            }
            
            if ( g_Plugin.pEditProcData && g_Plugin.pEditProcData->NextProc )
                lResult = g_Plugin.pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
            
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                if ( wParam == VK_SHIFT )
                {
                    OnEditSelKeyDown(hWnd, VK_SHIFT);
                }
                /*
                else if ( wParam == VK_LBUTTON )
                {
                    OnEditSelKeyDown(hWnd, VK_LBUTTON);
                }
                */
            }
            
            return lResult;
        }
        case WM_KEYUP:
        {
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                if ( wParam == VK_SHIFT )
                {
                    OnEditSelKeyUp(hWnd, VK_SHIFT);
                }
                /*
                else if ( wParam == VK_LBUTTON )
                {
                    OnEditSelKeyUp(hWnd, VK_LBUTTON);
                }
                */
            }
            break;
        }
        case WM_LBUTTONDOWN:
        {
            LRESULT lResult = 0;

            if ( g_Plugin.pEditProcData && g_Plugin.pEditProcData->NextProc )
                lResult = g_Plugin.pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
            
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                OnEditSelKeyDown(hWnd, VK_LBUTTON);
            }
            
            return lResult;
        }
        case WM_LBUTTONUP:
        {
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                OnEditSelKeyUp(hWnd, VK_LBUTTON);
            }
            break;
        }
        /*case WM_SETFOCUS:
        {
            if ( g_Plugin.bAkelEdit && (g_Plugin.dwInitializedMode & MODE_NOSELEOL) )
            {
                ShowCaret(hWnd);
            }
            break;
        }*/
        default:
        {
            break;
        }
    }
    
    if ( g_Plugin.pEditProcData && g_Plugin.pEditProcData->NextProc )
        return g_Plugin.pEditProcData->NextProc(hWnd, uMsg, wParam, lParam);
    else
        return 0;
}

LRESULT CALLBACK NewFrameProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch ( uMsg )
    {
        case WM_NOTIFY:
        {
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                LPNMHDR pnmh = (LPNMHDR) lParam;
                if ( pnmh->code == EN_SELCHANGE )
                {
                    EDITINFO ei;

                    SendMessage( g_Plugin.hMainWnd, 
                        AKD_GETEDITINFO, (WPARAM) NULL, (LPARAM) &ei );

                    if ( ei.hWndEdit == pnmh->hwndFrom )
                    {
                        OnEditSelChange(ei.hWndEdit);
                        /*ShowCaret(ei.hWndEdit);*/
                    }
                }
            }
            break;
        }
        case WM_MDIACTIVATE:
        {
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                clearSelectionState(&g_SelState);
            }
            break;
        }
        default:
        {
            break;
        }
    }
    
    CheckEditNotification(hWnd, uMsg, wParam, lParam);
    
    if ( g_Plugin.pFrameProcData && g_Plugin.pFrameProcData->NextProc )
        return g_Plugin.pFrameProcData->NextProc(hWnd, uMsg, wParam, lParam);
    else
        return 0;
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch ( uMsg )
    {
        case WM_NOTIFY:
        {
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                LPNMHDR pnmh = (LPNMHDR) lParam;
                if ( pnmh->code == EN_SELCHANGE )
                {
                    EDITINFO ei;

                    SendMessage( g_Plugin.hMainWnd, 
                        AKD_GETEDITINFO, (WPARAM) NULL, (LPARAM) &ei );

                    if ( ei.hWndEdit == pnmh->hwndFrom )
                    {
                        OnEditSelChange(ei.hWndEdit);
                        /*ShowCaret(ei.hWndEdit);*/
                    }
                }
            }
            break;
        }
        case WM_COMMAND:
        {
            if ( LOWORD(wParam) == IDM_NONMENU_DELLINE )
            {
                LRESULT lResult = 0;

                g_bDelLine = TRUE;
                
                if ( g_Plugin.pMainProcData && g_Plugin.pMainProcData->NextProc )
                    lResult = g_Plugin.pMainProcData->NextProc(hWnd, uMsg, wParam, lParam);

                g_bDelLine = FALSE;
                
                return lResult;
            }
            break;
        }
        case AKDN_EDIT_ONSTART:
        {
            /*
            // this commented code is already inside AkelPad
            DWORD dwEventMask;
            HWND  hEdit;
        
            hEdit = (HWND) wParam;
            dwEventMask = (DWORD) SendMessage(hEdit, EM_GETEVENTMASK, 0, 0);
            dwEventMask |= ENM_SELCHANGE;
            SendMessage(hEdit, EM_SETEVENTMASK, 0, dwEventMask);
            */
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                clearSelectionState(&g_SelState);
            }
            break;
        }
        case AKDN_EDIT_ONFINISH:
        {
            if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
            {
                clearSelectionState(&g_SelState);
            }
            break;
        }
        default:
        {
            break;
        }
    }
    
    CheckEditNotification(hWnd, uMsg, wParam, lParam);
    
    if ( g_Plugin.pMainProcData && g_Plugin.pMainProcData->NextProc )
        return g_Plugin.pMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
    else
        return 0;
}

void OnEditSelKeyDown(HWND hEdit, unsigned int uKey)
{
    if ( uKey == VK_SHIFT )
        g_SelState.bShiftDown = TRUE;
    else if ( uKey == VK_LBUTTON )
        g_SelState.bMouseLButtonDown = TRUE;

    g_SelState.SelKeyDownRange.cpMin = 0;
    g_SelState.SelKeyDownRange.cpMax = 0;
    SendMessage( hEdit, EM_EXGETSEL, 0, (LPARAM) &g_SelState.SelKeyDownRange );
}

void OnEditSelKeyUp(HWND hEdit, unsigned int uKey)
{
    BOOL bSelChanging = isSelectionChanging(&g_SelState);

    if ( uKey == VK_SHIFT )
        g_SelState.bShiftDown = FALSE;
    else if ( uKey == VK_LBUTTON )
        g_SelState.bMouseLButtonDown = FALSE;

    if ( bSelChanging && isSelectionChanged(&g_SelState) )
    {
        OnEditSelectionChanged(hEdit);
    }
}

void OnEditSelChange(HWND hEdit)
{
    if ( g_bDelLine )
        return;
    
    if ( !g_SelState.bSelJustChanged )
    {
        g_SelState.bSelChange = TRUE;

        if ( isSelectionChanged(&g_SelState) )
        {
            OnEditSelectionChanged(hEdit);
        }
    }
    else
    {
        g_SelState.bSelJustChanged = FALSE;
    }
}

void OnEditSelectionChanged(HWND hEdit)
{
    if ( !g_SelState.bColumnSelection )
    {
        CHARRANGE cr = { 0, 0 };
        
        SendMessage( hEdit, EM_EXGETSEL, 0, (LPARAM) &cr );
        if ( cr.cpMin != cr.cpMax )
        {
            if ( (cr.cpMax != g_SelState.SelKeyDownRange.cpMax) ||
                 (cr.cpMin >= g_SelState.SelKeyDownRange.cpMin) )
            {
                int     i;
                wchar_t szTextW[4];

                //if ( cr.cpMin == 0 )
                {
                    GETTEXTLENGTHEX gtl;
                    int nTextLen;

                    gtl.flags = GTL_DEFAULT | GTL_PRECISE;
                    gtl.codepage = 0;
                    nTextLen = (int) SendMessage(hEdit, EM_GETTEXTLENGTHEX, (WPARAM)&gtl, 0);

                    if ( cr.cpMax >= nTextLen )
                    {
                        // selection at EOF

                        g_SelState.bSelChange = FALSE;
                        g_SelState.SelKeyDownRange.cpMin = 0;
                        g_SelState.SelKeyDownRange.cpMax = 0;
                        return;
                    }
                }

                if ( g_Plugin.bOldWindows )
                {
                    TEXTRANGEA trA;

                    trA.chrg.cpMin = (cr.cpMax >= 2) ? (cr.cpMax - 2) : 0;
                    trA.chrg.cpMax = cr.cpMax;
                    trA.lpstrText = (LPSTR) szTextW;
                    i = (int) SendMessageA( hEdit, EM_GETTEXTRANGE, 0, (LPARAM) &trA );
                    if ( i < (trA.chrg.cpMax - trA.chrg.cpMin) )
                    {
                        // last symbol is '\0' - end of file
                        if ( cr.cpMax - cr.cpMin == 1 ) // just '\0' is selected
                            --cr.cpMax;
                    }
                    else
                    {
                        if ( (i > 0) && (trA.lpstrText[i-1] == '\n') )
                        {
                            --i;
                            --cr.cpMax;
                        }
                        if ( (i > 0) && (trA.lpstrText[i-1] == '\r') )
                        {
                            --cr.cpMax;
                        }
                    }
                    if ( cr.cpMax != trA.chrg.cpMax )
                    {
                        g_SelState.bSelJustChanged = TRUE;
                        SendMessage( hEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
                    }
                }
                else
                {
                    TEXTRANGEW trW;

                    trW.chrg.cpMin = (cr.cpMax >= 2) ? (cr.cpMax - 2) : 0;
                    trW.chrg.cpMax = cr.cpMax;
                    trW.lpstrText = (LPWSTR) szTextW;
                    i = (int) SendMessageW( hEdit, EM_GETTEXTRANGE, 0, (LPARAM) &trW );
                    if ( i < (trW.chrg.cpMax - trW.chrg.cpMin) )
                    {
                        // last symbol is L'\0' - end of file
                        if ( cr.cpMax - cr.cpMin == 1 ) // just L'\0' is selected
                            --cr.cpMax;
                    }
                    else
                    {
                        if ( (i > 0) && (trW.lpstrText[i-1] == L'\n') )
                        {
                            --i;
                            --cr.cpMax;
                        }
                        if ( (i > 0) && (trW.lpstrText[i-1] == L'\r') )
                        {
                            --cr.cpMax;
                        }
                    }
                    if ( cr.cpMax != trW.chrg.cpMax )
                    {
                        g_SelState.bSelJustChanged = TRUE;
                        SendMessage( hEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
                    }
                }
            }
        }
    }
    else
    {
        g_SelState.bColumnSelection = FALSE;
    }

    g_SelState.bSelChange = FALSE;
    g_SelState.SelKeyDownRange.cpMin = 0;
    g_SelState.SelKeyDownRange.cpMax = 0;
}

BOOL OnEditEndKeyDown(HWND hEdit, LPARAM lParam)
{
    int       nLine;
    int       nLineIndex;
    int       nLineLen;
    CHARRANGE cr = { 0, 0 };

    SendMessage(hEdit, EM_EXGETSEL, 0, (LPARAM) &cr);
    nLine = (int) SendMessage(hEdit, EM_EXLINEFROMCHAR, 0, (LPARAM)(-1));
    // wow, "magic" parameter (-1) gives REAL line number!!!
    nLineIndex = (int) SendMessage(hEdit, EM_LINEINDEX, nLine, 0);
    nLineLen = (int) SendMessage(hEdit, EM_LINELENGTH, nLineIndex, 0);
    if ( nLineLen <= 0 )
        return FALSE;

    /*
    if ( nLineIndex == cr.cpMax ) // line beginning or WordWrap'ed line end
    {
        CHARRANGE cr2 = { 0, 0 };
        
        HideCaret(hEdit);                                  // hide caret
        SendMessage(hEdit, WM_SETREDRAW, FALSE, 0);        // forbid redraw
        g_bProcessEndKey = FALSE;
        SendMessage(hEdit, WM_KEYDOWN, VK_END, lParam);    // emulate End pressed
        g_bProcessEndKey = TRUE;
        SendMessage(hEdit, EM_EXGETSEL, 0, (LPARAM) &cr2); // get new selection
        SendMessage(hEdit, EM_EXSETSEL, 0, (LPARAM) &cr);  // restore selection
        SendMessage(hEdit, WM_SETREDRAW, TRUE, 0);         // allow redraw
        ShowCaret(hEdit);                                  // show caret

        if ( cr.cpMax == cr2.cpMax ) // WordWrap'ed line end
        {
            --nLine; // previous line
            nLineIndex = (int) SendMessage(hEdit, EM_LINEINDEX, nLine, 0);
            nLineLen = (int) SendMessage(hEdit, EM_LINELENGTH, nLineIndex, 0);
            if ( nLineLen <= 0 )
                return FALSE;
        }
    }
    */
     
    if ( g_Plugin.dwInitializedMode & MODE_SMARTEND )
    {
        int  nSize;
        BOOL bShiftPressed = (GetKeyState(VK_SHIFT) & 0x80) ? TRUE : FALSE;

        if ( g_Plugin.bOldWindows )
            nSize = (nLineLen + 1)*sizeof(char);
        else
            nSize = (nLineLen + 1)*sizeof(wchar_t);
     
        if ( bufReserve(&g_TextBuf, nSize) )
        {
            int i;

            if ( g_Plugin.bOldWindows )
            {
                TEXTRANGEA trA;
                LPSTR      lpText = (LPSTR) g_TextBuf.pData;

                lpText[0] = 0;
                trA.chrg.cpMin = nLineIndex;
                trA.chrg.cpMax = nLineIndex + nLineLen;
                trA.lpstrText = lpText;
                i = (int) SendMessageA( hEdit, EM_GETTEXTRANGE, 0, (LPARAM) &trA );
                while ( --i >= 0 )
                {
                    if ( (lpText[i] != ' ') && (lpText[i] != '\t') )
                        break;
                }
            }
            else
            {
                TEXTRANGEW trW;
                LPWSTR     lpText = (LPWSTR) g_TextBuf.pData;

                lpText[0] = 0;
                trW.chrg.cpMin = nLineIndex;
                trW.chrg.cpMax = nLineIndex + nLineLen;
                trW.lpstrText = lpText;
                i = (int) SendMessageW( hEdit, EM_GETTEXTRANGE, 0, (LPARAM) &trW );
                while ( --i >= 0 )
                {
                    if ( (lpText[i] != L' ') && (lpText[i] != L'\t') )
                        break;
                }
            }
        
            ++i;
            
            if ( cr.cpMax == nLineIndex + nLineLen )
            {
                // selection ends at the end of line
                if ( (cr.cpMax == cr.cpMin) && (i == nLineLen) )
                {
                    // The caret is at the end of line (no selection).
                    // No trailing tabs or spaces, thus no SmartEnd available:
                    // there's nowhere to move.
                    return TRUE;
                }
                else if ( (cr.cpMax != cr.cpMin) && g_bSmartEndMode2 && !bShiftPressed )
                {
                    // altSmartEnd: jump to the end of line
                    return FALSE;
                }
            }
            
            if ( g_bSmartEndMode2 ? 
                  (cr.cpMax == nLineIndex + nLineLen) : 
                   (cr.cpMax != nLineIndex + i) 
               )
            {
                if ( bShiftPressed && (cr.cpMin == nLineIndex + i) )
                {
                    cr.cpMin = cr.cpMax;
                    cr.cpMax = nLineIndex + nLineLen;
                }
                else
                {
                    if ( !bShiftPressed )
                        cr.cpMin = nLineIndex + i;
                    cr.cpMax = nLineIndex + i;
                }
            }
            else
            {
                if ( !bShiftPressed )
                    cr.cpMin = nLineIndex + nLineLen;
                cr.cpMax = nLineIndex + nLineLen;
            }

            if ( cr.cpMax == nLineIndex + nLineLen )
            {
                if ( nLine < (int) SendMessage(hEdit, EM_EXLINEFROMCHAR, 0, cr.cpMax) )
                    return FALSE;
            }
            
            SendMessage( hEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
        }
    }
    else if ( g_Plugin.dwInitializedMode & MODE_NOSELEOL )
    {
        cr.cpMax = nLineIndex + nLineLen;
        SendMessage( hEdit, EM_EXSETSEL, 0, (LPARAM) &cr );
    }

    return TRUE;
}

void CheckEditNotification(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    if ( g_Plugin.bAkelEdit && (g_Plugin.dwInitializedMode & MODE_NOSELEOL) )
    {
        if ( (uMsg == WM_NOTIFY) && (wParam == ID_EDIT) )
        {
            if ( ((NMHDR *)lParam)->code == AEN_SELCHANGING )
            {
                AENSELCHANGE* psc = (AENSELCHANGE *) lParam;
                g_SelState.bColumnSelection = 
                  (psc->aes.dwFlags & AESELT_COLUMNON) ? TRUE : FALSE;
                /*if ( !g_SelState.bColumnSelection )
                {
                    HideCaret( ((NMHDR *)lParam)->hwndFrom );
                }*/
            }
        }
    }
}
